"""Tests for clementine.screenshot."""

import os
from unittest import mock

import pytest

from clementine import screenshot


def test_screenshot(tmp_path):
    webpage = tmp_path / "index.html"
    webpage.write_text(
        """
      <!DOCTYPE html>
      <html>
        <head>
          <style>
            body {
              background-color: #000;
              color: #fff;
            }
            @media (prefers-color-scheme: dark) {
              body {
                background-color: #fff;
                color: #000;
              }
            }
          </style>
        <head>
        <body>
          <h1>Hello, world!</h1>
        </body>
      </html>
      """
    )
    expected_paths = [
        os.path.join(tmp_path, "screenshot-light.png"),
        os.path.join(tmp_path, "screenshot-dark.png"),
    ]

    paths = screenshot.screenshot(f"file://{str(webpage)}", output_dir=tmp_path)
    assert sorted(paths) == sorted(expected_paths)
    for path in paths:
        assert os.path.exists(path)
        assert os.stat(path).st_size > 0


@mock.patch.dict("builtins.__dict__", {"__IPYTHON__": True})
def test_screenshot_raises_error_in_notebooks():
    with pytest.raises(RuntimeError) as execinfo:
        screenshot.screenshot("https://example.com")
        assert "Use `screenshot.async_screenshot()` instead." in str(execinfo.value)



"""Screenshot utilities."""

import asyncio
import os
from typing import Optional

from clementine import env, importing

__all__ = ["async_screenshot", "screenshot"]

_PLAYWRIGHT_IMPORT_ERROR_MESSAGE = (
    "The Playwright Python package is not installed. Install it with "
    "`pip install playwright`. Install Playwright's browser dependencies with "
    "`playwright install webkit`."
)

pw_async_api = importing.maybe_import_module(
    "playwright.async_api", _PLAYWRIGHT_IMPORT_ERROR_MESSAGE
)
pw_api_types = importing.maybe_import_module(
    "playwright._impl._api_types", _PLAYWRIGHT_IMPORT_ERROR_MESSAGE
)


async def async_screenshot(
    url: str,
    *,
    output_dir: Optional[str] = None,
    filename: Optional[str] = None,
    full_page: bool = False,
) -> list[str]:
    """Take screenshots of a webpage in light and dark mode.

    Use `screenshot.async_screenshot()` in asynchronous contexts, such as in a Jupyter
    notebook. Use `screenshot.screenshot()` in synchronous contexts, such as in a
    synchronous Python script.

    **Prerequisites**:

    1. Install the Playwright Python package: `pip install playwright`.
    2. Install the Playwright browser dependencies: `playwright install webkit`.


    **Example usage**:

    In a Python script:

    ```python
    from clementine import screenshot

    paths = screenshot.screenshot('https://example.com')
    ```

    In a Jupyter notebook or other asynchronous context:

    ```python
    from clementine import screenshot
    from IPython.display import display, Image

    paths = await screenshot.async_screenshot('https://example.com')
    for path in paths:
        display(Image(path, width=600))
    ```

    Args:
      url: The URL to take screenshots of. The URL must include the scheme, e.g.
        https:// or file:///.
      output_dir: The directory to save the screenshots to. Defaults to
        'screenshots' in the current working directory.
      filename: The base filename to use for the screenshots. Defaults to
        'screenshot'. The screenshots will be saved as screenshot-light.png and
        screenshot-dark.png, for example.
      full_page: Whether to take a screenshot of the full page. Defaults to False.

    Raises:
      ImportError: If the Playwright Python package is not installed.
      Error: If the browser cannot be launched or an invalid url is provided.

    Returns:
      A list of the paths to the screenshots.
    """
    output_dir = output_dir or os.path.join(os.getcwd(), "screenshots")
    filename = filename or "screenshot"
    output_paths = []

    async with pw_async_api.async_playwright() as p:
        try:
            browser = await p.webkit.launch()
        except pw_api_types.Error as e:
            raise OSError(
                "The browser could not be launched. Install the Playwright "
                "browser dependencies with `playwright install webkit`."
            ) from e

        page = await browser.new_page()
        await page.goto(url)

        os.makedirs(output_dir, exist_ok=True)
        for color_scheme in ("light", "dark"):
            await page.emulate_media(color_scheme=color_scheme)
            path = os.path.join(output_dir, f"{filename}-{color_scheme}.png")
            await page.screenshot(path=path, full_page=full_page)
            output_paths.append(path)

        await browser.close()

    return output_paths


def screenshot(
    url: str,
    *,
    output_dir: Optional[str] = None,
    filename: Optional[str] = None,
    full_page=False,
) -> list[str]:
    """Runs the coroutine function `screenshot.async_screenshot()` and returns its
    results."""
    screenshot.__doc__ = async_screenshot.__doc__
    if env.is_notebook():
        raise RuntimeError(
            "The function `screenshot.screenshot()` cannot be called in "
            "an asynchronous context, such as in a Jupyter notebook. "
            "Use `screenshot.async_screenshot()` instead."
        )
    return asyncio.run(
        async_screenshot(
            url, output_dir=output_dir, filename=filename, full_page=full_page
        )
    )



"""Importing utilities."""

import importlib
from types import ModuleType
from typing import Optional, Union

from clementine import errors

__all__ = ["MissingModulePlaceholder", "maybe_import_module"]


# Inspired by https://github.com/koaning/embetter/blob/main/embetter/error.py
class MissingModulePlaceholder:
    """A placeholder for a module that is not installed. This allows us to
    defer raising an error until the module is actually used.
    """

    def __init__(self, name: str, msg: Optional[str] = None):
        self._name = name
        self._msg = msg or (
            f"The package {name} is not installed. Install it with "
            f"`pip install {name}`."
        )

    def __getattr__(self, *args, **kwargs):
        """Raises an error when an attribute is accessed."""
        raise errors.NotInstalledError(self._name, self._msg)

    def __call__(self, *args, **kwargs):
        """Raises an error when the object is called."""
        raise errors.NotInstalledError(self._name, self._msg)


def maybe_import_module(
    name: str, error_message: Optional[str] = None
) -> Union[ModuleType, MissingModulePlaceholder]:
    """Tries to import a module by name and return it. If the module can't be
    imported, a placeholder object is returned instead.

    Args:
        name: The name of the module to import.
        error_message: An error message to pass to the placeholder object.
    """
    try:
        return importlib.import_module(name)
    except ImportError:
        return MissingModulePlaceholder(name, error_message)



"""Package dependencies for clementine."""

from collections import namedtuple
from dataclasses import dataclass
from typing import Union

__all__ = [
    "ClementineExtras",
    "Extra",
    "dependencies",
    "extras",
    "optional_dependencies",
    "python_version",
    "recommended_dependencies",
]

_VERSIONS = {
    "bandit": "bandit[toml]",
    "black": "black",
    "bokeh": "bokeh",
    "build": "build",
    "catppuccin-matplotlib": "catppuccin-matplotlib",
    "click": "click",
    "datasets": "datasets",
    "gradio": "gradio",
    "isort": "isort",
    "mkdocs-walt": "mkdocs-walt",
    "mkdocstrings": "mkdocstrings[python]",
    "mypy": "mypy",
    "numpy": "numpy",
    "notebook": "notebook",
    "pandas": "pandas",
    "pip-tools": "pip-tools",
    "playwright": "playwright",
    "pre-commit": "pre-commit",
    "pyperclip": "pyperclip",
    "pytest": "pytest",
    "rich": "rich",
    "ruff": "ruff",
    "scikit-learn": "scikit-learn",
    "spacy": "spacy",
    "tomlkit": "tomlkit",
    "transformers": "transformers",
    "twine": "twine",
    "wandb": "wandb",
}


@dataclass(frozen=True)
class Extra:
    """Basic information about an extra dependency."""

    name: str
    description: str
    deps: tuple[str, ...]


ClementineExtras = namedtuple(
    "ClementineExtras",
    ["blossom", "fruit", "leaves", "pulp", "rind", "seeds", "sprout", "tree"],
)


def _make_deps_list(package_names: Union[list[str], tuple[str, ...]]) -> list[str]:
    """Returns a list of dependencies with their constraints.

    Raises: ValueError if a `package_name` is not in the list of known dependencies.
    """
    deps = []
    for package_name in package_names:
        if package_name not in _VERSIONS:
            raise ValueError(
                f"Package '{package_name}' is not in the list of known dependencies: "
                f"{list(_VERSIONS.keys())}"
            )
        deps.append(_VERSIONS[package_name])
    return deps


_blossom = Extra(
    name="blossom",
    description="visualization and demonstration tools",
    deps=("bokeh", "gradio", "playwright"),
)

_fruit = Extra(
    name="fruit",
    description="machine learning libraries",
    deps=("catppuccin-matplotlib", "numpy", "pandas", "scikit-learn"),
)

_leaves = Extra(
    name="leaves",
    description="documentation tools",
    deps=("mkdocs-walt", "mkdocstrings"),
)

_pulp = Extra(
    name="pulp",
    description="auxiliary ml libraries",
    deps=("datasets", "spacy", "transformers", "wandb"),
)

_rind = Extra(
    name="rind",
    description="packaging tools",
    deps=("build", "tomlkit", "twine"),
)

_seeds = Extra(
    name="seeds",
    description="general development tools",
    deps=(
        "bandit",
        "black",
        "isort",
        "mypy",
        "pip-tools",
        "pre-commit",
        "pytest",
        "ruff",
    ),
)

_sprout = Extra(
    name="sprout",
    description="exploration and prototyping tools",
    deps=("notebook", "rich"),
)

_tree = Extra(
    name="tree",
    description="all optional dependencies",
    deps=(
        *_blossom.deps,
        *_fruit.deps,
        *_leaves.deps,
        *_pulp.deps,
        *_rind.deps,
        *_seeds.deps,
        *_sprout.deps,
    ),
)

extras = ClementineExtras(
    blossom=_blossom,
    fruit=_fruit,
    leaves=_leaves,
    pulp=_pulp,
    rind=_rind,
    seeds=_seeds,
    sprout=_sprout,
    tree=_tree,
)


def dependencies():
    """Returns a list of clementine's required dependencies."""
    return []


def optional_dependencies():
    """Returns a dictionary of clementine's optional dependencies."""
    return {
        "all": _make_deps_list(["click", "playwright", "pyperclip"]),
    }


def recommended_dependencies():
    """Returns a dictionary of recommended dependencies.

    The recommended dependencies are a collection of libraries and frameworks
    you might want to use when starting a new project. These dependencies aren't
    necessarily used in clementine, but they're made available as extra packages
    for easier installation."""
    return {extra.name: _make_deps_list(extra.deps) for extra in extras}


def python_version():
    """Returns clementine's required Python version."""
    return ">=3.9"



"""The clementine CLI."""

from clementine import about, importing
from clementine.cli import commands

click = importing.maybe_import_module("click")

_CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
}


@click.group(context_settings=_CONTEXT_SETTINGS, help=about.__description__)
@click.version_option(
    about.__version__, "-v", "--version", message=about.__pretty_version__
)
def main():
    """The main entry point for the clementine CLI."""
    pass


main.add_command(commands.info)
main.add_command(commands.screenshot)

if __name__ == "__main__":
    main()



"""The clementine CLI screenshot commands."""

from clementine import importing
from clementine.screenshot import screenshot as _screenshot

click = importing.maybe_import_module("click")


@click.command()
@click.argument("url")
@click.option(
    "-o",
    "--output-dir",
    help=(
        'The directory to save the screenshots to. Defaults to "screenshots" '
        "in the current working directory."
    ),
)
@click.option(
    "-f",
    "--filename",
    help=(
        'The base filename to use for the screenshots. Defaults to "screenshot". '
        "The screenshots will be saved as screenshot-light.png and "
        "screenshot-dark.png, for example."
    ),
)
@click.option(
    "--full-page/--no-full-page",
    default=False,
    help="Whether to take a screenshot of the full page. Defaults to False.",
)
def screenshot(url, output_dir, filename, full_page):
    """Take screenshots of a webpage in light and dark mode. The URL must include
    the scheme, e.g. https:// or file:///.
    """
    paths = _screenshot(
        url, output_dir=output_dir, filename=filename, full_page=full_page
    )
    click.echo(f"The screenshots of {url} were saved at these paths: ")
    for path in paths:
        click.echo(f"- {path}")



"""Tests for clementine.env."""

import importlib.machinery
import importlib.util
from unittest import mock

import pytest

from clementine import env, errors

_MODULE_NAME = "placeholder_module"
_MODULE_SPEC = importlib.machinery.ModuleSpec(_MODULE_NAME, object())  # type: ignore
_MODULES_DICT = {_MODULE_NAME: importlib.util.module_from_spec(_MODULE_SPEC)}
_MODULES_DICT_WITH_MISSING_PACKAGE = {_MODULE_NAME: None}


@mock.patch.dict("sys.modules", _MODULES_DICT)
def test_is_installed():
    assert env.is_installed(_MODULE_NAME)


@mock.patch.dict("sys.modules", _MODULES_DICT_WITH_MISSING_PACKAGE)
def test_is_installed_with_missing_package():
    assert not env.is_installed(_MODULE_NAME)


@mock.patch.dict("sys.modules", _MODULES_DICT)
def test_ensure_is_installed():
    env.ensure_is_installed(_MODULE_NAME)


@mock.patch.dict("sys.modules", _MODULES_DICT_WITH_MISSING_PACKAGE)
def test_ensure_is_install_fails_with_missing_package():
    with pytest.raises(errors.NotInstalledError) as execinfo:
        env.ensure_is_installed(_MODULE_NAME)
        assert f"'{_MODULE_NAME}' is not installed" in str(execinfo.value)



"""Environment utilities."""

import builtins
import importlib.metadata
import importlib.util
import platform
from typing import Optional

from clementine import errors

__all__ = ["ensure_is_installed", "info", "is_installed", "is_notebook"]


def is_installed(package: str) -> bool:
    """Returns True if a package is installed or is a built-in."""
    return importlib.util.find_spec(package) is not None


def ensure_is_installed(package: str, message: Optional[str] = None):
    """Raises a NotInstalledError if a package is not installed."""
    if not is_installed(package):
        raise errors.NotInstalledError(package, message)


def info(packages: Optional[list[str]] = None) -> dict[str, str]:
    """Returns package installation and environment information."""
    env = {
        "Python version": platform.python_version(),
        "Platform": platform.platform(),
    }

    if packages:
        for package in packages:
            try:
                env[f"{package} version"] = importlib.metadata.version(package)
            except importlib.metadata.PackageNotFoundError:
                env[f"{package} version"] = "Not installed."

    return env


def is_notebook() -> bool:
    """Returns True if the current environment is a Jupyter notebook."""
    return getattr(builtins, "__IPYTHON__", False)



"""Tests for clementine.importing."""

import pytest

from clementine import errors, importing


def test_missing_module_placeholder():
    name = "not_installed"
    error_message = f"{name} is not installed. Using a placeholder instead."
    placeholder = importing.MissingModulePlaceholder(name)
    with pytest.raises(errors.NotInstalledError) as execinfo:
        placeholder.attribute
        assert str(execinfo.value) == error_message
    with pytest.raises(errors.NotInstalledError) as execinfo:
        placeholder()
        assert str(execinfo.value) == error_message



"""The clementine CLI info commands."""

from clementine import about, importing
from clementine.env import info as _info

click = importing.maybe_import_module("click")
pyperclip = importing.maybe_import_module("pyperclip")


@click.command()
@click.option(
    "--clipboard/--no-clipboard",
    default=False,
    help="Whether to copy the information to the clipboard. Defaults to False.",
)
def info(clipboard):
    """Show information about your environment and clementine installation."""
    env = "\n".join([f"{k}: {v}" for k, v in _info([about.__title__]).items()])
    if clipboard:
        pyperclip.copy(env)
    click.echo(env)



"""Some metadata for the clementine package."""

import webbrowser

__title__ = "clementine"
__description__ = "🍊 clementine is a sweet little Python package."
__author__ = "Suzen Fylke"
__version__ = "2023.8.24"  # Calver with format YYYY.MM.DD
__pretty_version__ = f"{__title__} v{__version__}"
__documentation__ = "https://codesue.github.io/clementine"
__repository__ = "https://github.com/codesue/clementine"


def docs():
    """Opens the documentation website in a browser."""
    webbrowser.open(__documentation__)


def repo():
    """Opens the repository website in a browser."""
    webbrowser.open(__repository__)



"""Errors."""

from typing import Optional

__all__ = ["NotInstalledError"]


class NotInstalledError(ImportError):
    """Raised when a package is not installed."""

    def __init__(self, name: str, msg: Optional[str] = None):
        self.name = name
        self.msg = msg or f"The package '{name}' is not installed."
